/*
	General libsharedmime API

	This is the implementation of libsharedmime's public API.
	
	Author: Matthew Allen <fret@memecode.com>
	Date: 5/12/2003
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "mime-types.h"
#include "mime-types-priv.h"
#include "glob.h"
#include "magic.h"

// What goes in here?
const char *mimetypes_identity = "";
const char *mimetypes_version = "";

class SharedMime
{
public:
	GlobLut Globs[3];
	MagicLut Magic[3];
	char *BasePaths[3];
	char *DefaultType;
	bool Inited;

	SharedMime()
	{
		Inited = false;
		BasePaths[0] = MIME_DIR_USER;
		BasePaths[1] = MIME_DIR_LOCAL;
		BasePaths[2] = MIME_DIR_GLOBAL;
		DefaultType = NewStr("application/x-not-recognised");
	}
	
	~SharedMime()
	{
		DeleteArray(DefaultType);
	}
	
	void Init()
	{
		if (!Inited)
		{
			for (int i=0; i<CountOf(BasePaths); i++)
			{
				char p[256];
				if (MakePath(p, BasePaths[i], "globs"))
				{
					Globs[i].Load(p);
				}
				if (MakePath(p, BasePaths[i], "magic"))
				{
					Magic[i].Load(p);
				}
			}
			
			Inited = true;
		}		
	}
	
	char *GetPathMimeType(char *Path, bool *IsDir)
	{
		struct stat s;
		
		lstat(Path, &s);
		
		if (S_ISDIR(s.st_mode))
		{
			if (IsDir) *IsDir = true;
			return "inode/directory";
		}
		else if (S_ISLNK(s.st_mode))
		{
			return "inode/symlink";
		}
		else if (S_ISSOCK(s.st_mode))
		{
			return "inode/socket";
		}
		else if (S_ISBLK(s.st_mode))
		{
			return "inode/blockdevice";
		}
		else if (S_ISCHR(s.st_mode))
		{
			return "inode/chardevice";
		}
		else
		{
			for (int i=0; i<CountOf(Globs); i++)
			{
				char *m = Globs[i].Lookup((char*) Path);
				if (m)
				{
					static char type[256];
					strcpy(type, m);
					DeleteArray(m);
					return type;
				}
			}
		}

		return 0;
	}
	
	char *GetDataMimeType(char *Data, int Len, char *Hint, bool Debug = false)
	{
		for (int i=0; i<CountOf(BasePaths); i++)
		{
			char *m = Magic[i].GetType(Data, Len, Hint, Debug);
			if (m)
			{
				return m;
			}
		}
		
		return 0;		
	}
	
	bool IsTextFile(char *Buf, int Len)
	{
		for (unsigned char *s = (unsigned char *) Buf; s && *s && Len-- > 0; s++)
		{
			if (*s != '\t' &&
				*s != '\n' &&
				*s != '\r' &&
				*s != 12 && // form feed
				*s < ' ')
			{
				return false;
			}
		}
		
		return true;
	}
	
	char *GetContentMimeType(char *Path, char *Hint)
	{
		char *MimeType = 0;
		
		bool Debug = false;

		int f = open(Path, O_RDONLY);
		if (f >= 0)
		{
			int BufLen = 64 << 10;
			char *Buf = new char[BufLen];
			if (Buf)
			{
				BufLen = read(f, Buf, BufLen);
				if (BufLen > 0)
				{
					MimeType = GetDataMimeType(Buf, BufLen, Hint, Debug);
					if (! MimeType && IsTextFile(Buf, BufLen))
					{
						MimeType = NewStr("text/plain");
					}
				}
				else
				{
					// printf("%s:%i - Couldn't read from '%s'\n", __FILE__, __LINE__, Path);
				}
				
				DeleteArray(Buf);
			}
			
			close(f);
		}
		
		return MimeType;
	}
	
} Lib;

void
mimetypes_init()
{
	Lib.Init();
}

CONST_RETURN char *
mimetypes_set_default_type(const char *default_type)
{
	DeleteArray(Lib.DefaultType);
	if (default_type)
	{
		Lib.DefaultType = NewStr((char*)default_type);
	}	
	return Lib.DefaultType;
}

CONST_RETURN char *
mimetypes_get_file_type(const char *pathname, mimetypes_flags flags)
{
	if (pathname)
	{
		bool IsDir = false;
		char *PathMimeType = 0;
		char *ContentMimeType = 0;
		char Path[300];
		strcpy(Path, (char*)pathname);

		if (flags & MIMETYPES_CHECK_PATH)
		{
			PathMimeType = Lib.GetPathMimeType(Path, &IsDir);
		}
		// printf("Mime? '%s'\n\tPathMimeType=%s\n", Path, PathMimeType);
		
		if ((flags & MIMETYPES_CHECK_CONTENTS) != 0 && !IsDir)
		{
			ContentMimeType = Lib.GetContentMimeType(Path, PathMimeType);
		}
		// printf("\tContentMimeType=%s\n", ContentMimeType);

		static char MimeType[256];
		
		if (ContentMimeType)
		{
			if (stricmp(ContentMimeType, "text/xml") == 0 &&
				PathMimeType)
			{
				strcpy(MimeType, PathMimeType);
			}
			else
			{
				strcpy(MimeType, ContentMimeType);
			}
		}
		else if (PathMimeType)
		{
			strcpy(MimeType, PathMimeType);
		}
		else if (Lib.DefaultType)
		{
			strcpy(MimeType, Lib.DefaultType);
		}

		DeleteArray(ContentMimeType);

		#ifdef _DEBUG
		if (stricmp(MimeType, "application/x-compressed-tar") == 0)
		{
			strcpy(MimeType, "application/x-tgz");
		}
		else if (stricmp(MimeType, "application/x-bzip-compressed-tar") == 0)
		{
			strcpy(MimeType, "application/x-tbz");
		}
		#endif

		return MimeType[0] ? MimeType : 0;
	}
	
	return 0;
}

CONST_RETURN char *
mimetypes_get_data_type(const void *data, const int length)
{
	if (data && length > 0)
	{
		return Lib.GetDataMimeType((char*)data, length, 0);
	}
	
	return 0;
}

bool
mimetypes_decode(const char *type, char **media_type, char **sub_type)
{
	if (type)
	{
		static char buf[256];
		strcpy(buf, type);
		char *s = strchr(buf, '/');
		if (s)
		{
			*s++ = 0;
			
			if (media_type)
			{
				*media_type = buf;
			}			
			if (sub_type)
			{
				*sub_type = s;
			}
			
			return media_type != 0 || sub_type != 0;
		}
	}

	return false;
}

char *
mimetypes_convert_filename(const char *pathname, const char *mime_type)
{
	return 0;
}

bool
mimetypes_add_mime_dir(const char *path, bool high_priority, bool rescan)
{
	return false;
}

CONST_RETURN char *
mimetypes_get_type_info(const char *mime_type, const char *lang)
{
	return 0;
}
